//
//  MPQArchive.h
//  MPQ Framework
//
//  Created by BahamutZERO on Tue Oct 01 2002.
//  Copyright (c) 2002 MacStorm. All rights reserved.
//

/*!
  @header MPQArchive.h
  This header is automatically included by MPQ.h and provides the definition of the class 
  MPQArchive.
*/

@class MPQFile;

/*!
 	@class MPQArchive
	@abstract MPQArchive is the principal class of the MPQ Framework and represents an MPQ archive 
        document.
	@discussion MPQArchive is the principal class of the MPQ Framework and represents an MPQ archive
        document. There is a one-to-one relationship between MPQArchive instances and MPQ archives 
        on disks. You may create a new MPQArchive either with only a file limit parameter, 
        and provide the location of the archive on disk when saving, or create a new MPQArchive 
        with a path and a file limit parameter, in which case the file on disk is created right away.
        The first method better matches the Cocoa document model. Unlike most other MPQ libraries, 
        MPQArchive has deferred operations, meaning the file on disk is not modified until you save.
        Again, this behavior better matches the Cocoa document model.
*/
@interface MPQArchive : NSObject
{
	BOOL m_bIsOpen;
	BOOL m_bIsModified;
	BOOL m_bIsReadOnly;
	BOOL m_bSaveListfile;
	
	int m_archivefd;
	NSString *m_archivePath;
    
    DWORD m_dwCompressor;

	DWORD m_dwMPQOffset;
	DWORD m_dwInsertPoint;
	DWORD m_dwBlockSize;
	DWORD m_dwHashTableEntries;
	DWORD m_dwFileTableEntries;

	DWORD m_dwHashTableUsed;
	DWORD m_dwFileTableUsed;
    
    DWORD m_dwNumOpenFiles;
	DWORD *m_lpdwFileRefCount;

	LPBYTE m_lpbyReadBuffer;
	LPBYTE m_lpbyCompBuffer;

	struct HASHTABLEENTRY *m_lpvHashTable;
	struct FILETABLEENTRY *m_lpvFileTable;
    NSMutableArray **m_defer_data_table;
    DWORD m_numOps;
	LPSTR *m_ppszNameTable;
	
	NSMutableData **blockTableCache;

	DWORD * m_cryptKeyTable;
    
    id m_delegate;
    
    NSData *m_headerData;
	
	NSArray *m_metaAttribute;
	NSMutableData *m_attributeData;
}

/*! 
    @method archiveWithFileLimit:header:
	@abstract Creates and returns an autoreleased MPQArchive instance initialized with the given
        file limit. No file is created on disk.
	@discussion This is the recommended way of loading an existing archive from disk. Depending on where it is 
		stored and what permissions the archive file has, you may or may not be able to modify the archive.
	@param limit  An integer indicating the maximum number of files the new archive may
        contain. Minimum value is 16, maximum value is 0x80000. If you pass 0, MPQArchive will
        set this parameter to the default value of 1024. The file limit must be a power of 2 and will
        automatically be rounded off to one, so you may safely pass any integer.
    @param headerData  An NSData instance containing bytes that will be saved before the archive 
        at save time. Note that MPQ archives must be embedded at 512 bytes offsets, so the 
        data that you pass will be padded to the nearest multiple of 512. Header data 
        cannot be changed beyond initialization because of offset-adjusted files.
	@result  Returns the newly initialized MPQArchive object or nil on error.
*/
+ (id)archiveWithFileLimit:(unsigned)limit header:(NSData *)headerData;

/*! 
    @method archiveWithPath:
	@abstract Creates and returns an autoreleased MPQArchive instance initialized with the archive at 
        the specified path.
    @discussion Creates and returns an MPQArchive instance initialized with the archive at 
        the specified path.
	@param mpqPath  The POSIX path to the archive. Must be a fully expanded path. Must not be nil.
	@result  Returns the newly initialized MPQArchive object or nil on error.
*/
+ (id)archiveWithPath:(NSString *)mpqPath;

/*! 
    @method initWithFileLimit:header:
	@abstract Creates and returns an MPQArchive instance initialized with the given file 
		limit and header data. No file is created on disk.
	@discussion This is the core way of loading an existing archive from disk. Depending on where it is 
		stored and what permissions the archive file has, you may or may not be able to modify the archive.
	@param limit  An integer indicating the maximum number of files the new archive may 
        contain. Minimum value is 16, maximum value is 0x80000. If you pass 0, MPQArchive will 
        set this parameter to the default value of 1024. The file limit must be a power of 2 and will 
        automatically be rounded off to one, so you may safely pass any integer.
    @param headerData  An NSData instance containing bytes that will be saved before the archive 
        at save time. Note that MPQ archives must be embedded at 512 bytes offsets, so the 
        data that you pass will be padded to the nearest multiple of 512. Header data 
        cannot be changed beyond initialization because of offset-adjusted files.
	@result  Returns the newly initialized MPQArchive object or nil on error.
*/
- (id)initWithFileLimit:(unsigned)limit header:(NSData *)headerData;

/*! 
    @method initWithPath:
	@abstract Creates and returns an MPQArchive instance initialized with the archive at 
        the specified path.
    @discussion Creates and returns an MPQArchive instance initialized with the archive at 
        the specified path.
	@param mpqPath  The POSIX path to the archive. Must be a fully expanded path. Must not be nil.
	@result  Returns the newly initialized MPQArchive object or nil on error.
*/
- (id)initWithPath:(NSString *)mpqPath;

/*!
    @method delegate
    @abstract Returns the object's current delegate
	@discussion MPQArchive principally notifies its delegate about operations that are about to be performed or 
		that finished performing, and also to control operations (allow or disallow).
    @result  The object's delegate or nil.
*/
- (id)delegate;

/*!
    @method setDelegate:
    @abstract Set the archive's delegate object.
	@discussion MPQArchive principally notifies its delegate about operations that are about to be performed or 
		that finished performing, and also to control operations (allow or disallow).
*/
- (void)setDelegate:(id)anObject;

/*! 
    @method modified
	@abstract Returns the current document state of the archive.
	@discussion This will always return NO if the archive is read-only. When an archive was just opened, this should always 
		return NO.
	@result  Returns YES if the archive has been modified and NO if it has not.
*/
- (BOOL)modified;

/*! 
    @method readOnly
	@abstract Returns the read-only state of the archive.
	@discussion This will always return YES for new archives. This may return NO for existing archives depending on what permissions 
		the archive file has and on what type of media it is stored.
	@result  Returns YES if the archive is read-only and NO if it is read-write.
*/
- (BOOL)readOnly;

/*! 
    @method openFiles
	@abstract Returns the number of open MPQ files.
    @discussion There must be no open files prior to saving the archive.
	@result  Returns the number of open MPQ files as an integer.
*/
- (unsigned long)openFiles;

/*! 
    @method fileOpenCount:
	@abstract Returns the number of MPQFile instances for a given MPQ file.
    @discussion There must be no open files prior to saving the archive.
    @param position  The hash table position of the desired file. May be obtained 
        by the information methods of MPQArchive and MPQFile.
	@result  Returns the number of MPQFile instances for a given MPQ file.
*/
- (unsigned long)fileOpenCount:(unsigned long)position;

/*! 
    @method fileCount
	@abstract Returns the number of used and deleted files inside the archive.
    @discussion You may calculate the number of deleted files by substracting this number 
        with the result of usedFileCount.
	@result  Returns the number of used and deleted files inside the archive as an integer.
*/
- (unsigned long)fileCount;

/*! 
    @method usedFileCount
	@abstract Returns the number of used files inside the archive.
	@discussion This includes every file, even if the file name is not known.
	@result  Returns the number of used files inside the archive as an integer.
*/
- (unsigned long)usedFileCount;

/*! 
    @method maxFileCount
	@abstract Returns the maximum number of files the archive may contain.
	@discussion There is no way to augment this limit past the creation of an archive. Ergo if an archive is full, 
		you may have to delete some files or create a bigger archive and copy every file from one to the other, which you 
		may not be able to do if you don't know the file name of every file. This limit is inherent to the MPQ format and 
		not this framework.
	@result  Returns the maximum number of files the archive may contain as an integer.
*/
- (unsigned long)maxFileCount;

/*
    @method operationsCount
    @abstract Returns the number of operations that will be performed on the archive when it is saved.
    @discussion You can expect to receive as many save-time delegate messages reporting on what
        operation the framework is about to do or just completed as the result of this function.
    @result  Returns the number of operations as an integer.
*/
- (unsigned long)operationsCount;

/*! 
    @method path
	@abstract Returns the path of the archive associated with the instance.
    @discussion May be nil if the instance was initialized with archiveWithFileLimit: or 
        initWithFileLimit: and the archive was never saved yet.
	@result  Returns the absolute path of the archive as an integer.
*/
- (NSString *)path;

/*! 
    @method writeToFile:atomically:
	@abstract Saves the archive at path.
    @discussion Attempts to save the archive at path. If path is the same 
        as the path that was used to initialize the instance, the archive file is modified directly. 
        If path is different than the path used to initialize the instance, a new file is created 
        at path and the archive is saved there (save-as operation). Furthermore, the archive path of 
        the instance is changed to the new path and the old archive is closed.
        
        If the archive was not initialized with a path, a new file is created at path and the path of the 
        instance is set to path.
        
        In all cases, setting flag to YES makes the instance close any 
        currently open archive and create a new temporary archive file which will be 
        moved to path after saving is done. Note that the temporary file is actually a 
        copy of the original archive file, so if the archive is embedded in some 
        other file, all data not belonging to the archive is preserved.
        
        In all cases, if the method returns NO, the instance will be exactly like it was prior to the save 
        attempt.
	@param path  The location where to save the archive. Must not be nil.
    @param flag  If flag is YES, the instance closes any currently open archive and makes a new temporary 
        archive which is used for the save operation and is moved to path after saving is done.
	@result  Returns YES if saving was sucessful or NO if an error occured. If flag was YES, then the original 
        archive was reopened and the path of the instance set to the path of the original. The archive is 
        marked as not modified if this method succeeds. 
*/
- (BOOL)writeToFile:(NSString *)path atomically:(BOOL)flag;

/*! 
    @method storesListfile
	@abstract Returns whether or not the instance will save the list of files inside the archive in
        the internal MPQ listfile at save time.
    @discussion Each MPQArchive instance has a list of the paths of the files inside its associated 
        MPQ archive. Because MPQs don't inherently store the list of files they contain, MPQ editors have gone 
        around this by manually managing a file inside MPQ archives called (listfile).
        
        There are several advantages to having the path of a file, such as being able to re-compress it and/or re-encrypt it in 
        operations such as compaction. Without the path, it would be quasi-impossible to perform those operations. 
        Note that whenever you try to open a file of the archive, the MPQArchive instance will add the path 
        you used to open the file to its internal list of files if the file is actually found. You may also 
        add external listfiles (available from the web) to the instance's internal list of files using 
        the addListfileToFileList: and addArrayToFileList: methods.
        
        The default behavior is to save the listfile at save time.
	@result  Returns YES if the listfile will be added to the archive or NO if it will not.
*/
- (BOOL)storesListfile;

/*! 
    @method setStoresListfile:
	@abstract Set whether or not the instance will save the list of files inside the archive in
        the internal MPQ listfile at save time.
    @discussion Each MPQArchive instance has a list of the paths of the files inside its associated 
        MPQ archive. Because MPQs don't inherently store the list of files they contain, MPQ editors have gone 
        around this by manually managing a file inside MPQ archives called (listfile).
        
        There are several advantages to having the path of a file, such as being able to re-compress it and/or re-encrypt it in 
        operations such as compaction. Without the path, it would be quasi-impossible to perform those operations. 
        Note that whenever you try to open a file of the archive, the MPQArchive instance will add the path 
        you used to open the file to its internal list of files if the file is actually found. You may also 
        add external listfiles (available from the web) to the instance's internal list of files using 
        the addListfileToFileList: and addArrayToFileList: methods.
        
        It is recommended to save the listfile.
    @param bStore  YES to make the instance save the listfile at save time or NO to prevent it from doing so.
*/
- (void)setStoresListfile:(BOOL)bStore;

/*! 
    @method loadListfile
	@abstract Loads the internal MPQ listfile into the instance's own internal list of files. Helps to get valid file
        paths when using the file information routines.
    @discussion Each MPQArchive instance has a list of the paths of the files inside its associated 
        MPQ archive. Because MPQs don't inherently store the list of files they contain, MPQ editors have gone 
        around this by manually managing a file inside MPQ archives called (listfile).
        
        There are several advantages to having the path of a file, such as being able to re-compress it and/or re-encrypt it in 
        operations such as compaction. Without the path, it would be quasi-impossible to perform those operations. 
        Note that whenever you try to open a file of the archive, the MPQArchive instance will add the path 
        you used to open the file to its internal list of files if the file is actually found. You may also 
        add external listfiles (available from the web) to the instance's internal list of files using 
        the addListfileToFileList: and addArrayToFileList: methods.
        
        This method attemps to read the archive's (listfile) and add its content to the instance's internal list 
        of files. MPQArchive will make sure each entry in the listfile actually exists inside the archive 
        before adding it to the list.
*/
- (void)loadListfile;

/*! 
    @method compressor
	@abstract Returns the current default compressor of the instance.
	@discussion The default compressor is zlib.
	@result  A MPQCompressionType constant. See the header MPQSharedConstants.h for a list of valid compressor constants.
*/
- (MPQCompressionType)compressor;

/*! 
    @method setCompressor:
	@abstract Sets the default compressor of the instance.
	@discussion Generally speaking, DCL is more backward-compatible while zlib offers better speed and compression.
    @param compressor  An MPQCompressionType constant. May only be MPQZLIBCompression, 
        MPQDCLCompression or MPQDiabloCompression.
    @result  YES on sucess or NO on failure.
*/
- (BOOL)setCompressor:(MPQCompressionType)compressor;

/*! 
    @method header
	@abstract Returns the archive's header data if any.
    @discussion MPQ archives can be embedded in other files at 512 bytes padded addresses. This function will 
        return any data that was found before the archive when it was loaded.
	@result  An NSData containing the header data or nil if there was no header data.
*/
- (NSData *)header;

/*! 
    @method addArrayToFileList:
	@abstract Adds a list of paths to the instance's internal list of files.
    @discussion Each MPQArchive instance has a list of the paths of the files inside its associated 
        MPQ archive. Because MPQs don't inherently store the list of files they contain, MPQ editors have gone 
        around this by manually managing a file inside MPQ archives called (listfile).
        
        There are several advantages to having the path of a file, such as being able to re-compress it and/or re-encrypt it in 
        operations such as compaction. Without the path, it would be quasi-impossible to perform those operations. 
        Note that whenever you try to open a file of the archive, the MPQArchive instance will add the path 
        you used to open the file to its internal list of files if the file is actually found. You may also 
        add external listfiles (available from the web) to the instance's internal list of files using 
        the addListfileToFileList: and addArrayToFileList: methods.
        
        This method attemps to add each entry of the provided array to the instance's internal list 
        of files. MPQArchive will make sure each entry actually exists inside the archive 
        before adding it to the list.
    @param pListFile  A NSArray containing MPQ file paths. Note that the path separator MUST BE \, and not /.
    @result  YES on sucess or NO on failure.
*/
- (BOOL)addArrayToFileList:(NSArray *)pListFile;

/*! 
    @method addListfileToFileList:
	@abstract Adds an external listfile to the instance's internal list of files.
    @discussion Each MPQArchive instance has a list of the paths of the files inside its associated 
        MPQ archive. Because MPQs don't inherently store the list of files they contain, MPQ editors have gone 
        around this by manually managing a file inside MPQ archives called (listfile).
        
        There are several advantages to having the path of a file, such as being able to re-compress it and/or re-encrypt it in 
        operations such as compaction. Without the path, it would be quasi-impossible to perform those operations. 
        Note that whenever you try to open a file of the archive, the MPQArchive instance will add the path 
        you used to open the file to its internal list of files if the file is actually found. You may also 
        add external listfiles (available from the web) to the instance's internal list of files using 
        the addListfileToFileList: and addArrayToFileList: methods.
        
        This method attemps to add each entry of the provided listfile to the instance's internal list 
        of files. MPQArchive will make sure each entry actually exists inside the archive 
        before adding it to the list.
    @param listfile  A string containing the content of the listfile in the NSASCIIStringEncoding encoding. 
        Note that the path separator MUST BE \, and not /, and that the line breaks MUST BE \r\n and not \n or \r.
    @result  YES on sucess or NO on failure.
*/
- (BOOL)addListfileToFileList:(NSString *)listfile;

/*! 
    @method fileList
	@abstract Returns the instance's internal list of files.
    @discussion Each MPQArchive instance has a list of the paths of the files inside the 
        MPQ archive. Because MPQs don't inherently store the list of files they contain, MPQ editors have gone 
        around this by manually managing a file inside MPQ archives called (listfile).
        
        There are several advantages to having the path of a file, such as being able to re-compress it and/or re-encrypt it in 
        operations such as compaction. Without the path, it would be quasi-impossible to perform those operations. 
        Note that whenever you try to open a file of the archive, the MPQArchive instance will add the path 
        you used to open the file to its internal list of files if the file is actually found. You may also 
        add external listfiles (available from the web) to the instance's internal list of files using 
        the addListfileToFileList: and addArrayToFileList: methods.
    @result  An NSArray containing all the known paths of the instance's files. Nil on failure.
*/
- (NSArray *)fileList;

/*! 
    @method fileListWithFilter:
	@abstract Returns the instance's internal list of files filtered by a regular expression.
    @discussion Each MPQArchive instance has a list of the paths of the files inside the 
        MPQ archive. Because MPQs don't inherently store the list of files they contain, MPQ editors have gone 
        around this by manually managing a file inside MPQ archives called (listfile).
        
        There are several advantages to having the path of a file, such as being able to re-compress it and/or re-encrypt it in 
        operations such as compaction. Without the path, it would be quasi-impossible to perform those operations. 
        Note that whenever you try to open a file of the archive, the MPQArchive instance will add the path 
        you used to open the file to its internal list of files if the file is actually found. You may also 
        add external listfiles (available from the web) to the instance's internal list of files using 
        the addListfileToFileList: and addArrayToFileList: methods.
        
        WARNING: CURRENTLY NOT IMPLEMENTED. WILL SIMPLY RETURN THE RESULT OF fileList.
    @param filter  The regular expression upon which the known paths will be tested.
    @result  An NSArray containing all the known paths of the instance's files matching the expression. Nil on failure.
*/
- (NSArray *)fileListWithFilter:(NSString *)filter;

/*! 
    @method fileInfoEnumerator
	@abstract Returns an enumerator of file information dictionaries.
    @discussion File information dictionaries will always contain the following keys:
        
        * MPQFileLength: The size of the file in bytes as an integer.
        
        * MPQFileCompressedLength: The compressed size of the file in bytes as an integer.
        
        * MPQFileFlags: The file's flags as an integer. See the MPQFileFlag enum in 
        MPQSharedConstants.h for a list of valid bit values.
        
        * MPQFileLocale: The file's locale code as an integer. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
        
        * MPQFileHashA: The file's A hash as an integer. Mostly useless.
        
        * MPQFileHashB: The file's B hash as an integer. Mostly useless as well.
        
        * MPQFileHashPosition: The file's hash table position. Can be used as a unique key for that 
        particular file (in fact, a file's path is not a unique key, so this is the only unique key 
        for MPQ files).
        
        File information dictionaries may also contain the following key:
        
        * MPQFileName: The file's path inside the MPQ archive. Note that the path separator is \.
    @result  An NSEnumerator subclass object you can use to cycle through the file information dictionaries. Nil on failure.
*/
- (NSEnumerator *)fileInfoEnumerator;

/*!
	@method fileInfoForPosition:
	@abstract Returns the information dictionary for the file at the specified position
	@discussion File information dictionaries will always contain the following keys:
        
        * MPQFileLength: The size of the file in bytes as an integer.
        
        * MPQFileCompressedLength: The compressed size of the file in bytes as an integer.
        
        * MPQFileFlags: The file's flags as an integer. See the MPQFileFlag enum in 
        MPQSharedConstants.h for a list of valid bit values.
        
        * MPQFileLocale: The file's locale code as an integer. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
        
        * MPQFileHashA: The file's A hash as an integer. Mostly useless.
        
        * MPQFileHashB: The file's B hash as an integer. Mostly useless as well.
        
        * MPQFileHashPosition: The file's hash table position. Can be used as a unique key for that 
        particular file (in fact, a file's path is not a unique key, so this is the only unique key 
        for MPQ files).
        
        File information dictionaries may also contain the following key:
        
        * MPQFileName: The file's path inside the MPQ archive. Note that the path separator is \.
		
		Additionally, file information dictionaries may contain one or more MPQ file attribute keys.
		Please refer to the documentation of the header MPQSharedConstants.h for a list of valid keys.
	@param position  An integer specifying the position of the file you wish information about. Must not be out of bounds.
	@result  An NSDictionary object containing the file's information. Nil on failure.
*/
- (NSDictionary *)fileInfoForPosition:(unsigned)position;

/*! 
    @method fileInfoForFile:locale:
	@abstract Returns the information dictionary for the specified file.
    @discussion File information dictionaries will always contain the following keys:
        
        * MPQFileLength: The size of the file in bytes as an integer.
        
        * MPQFileCompressedLength: The compressed size of the file in bytes as an integer.
        
        * MPQFileFlags: The file's flags as an integer. See the MPQFileFlag enum in 
        MPQSharedConstants.h for a list of valid bit values.
        
        * MPQFileLocale: The file's locale code as an integer. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
        
        * MPQFileHashA: The file's A hash as an integer. Mostly useless.
        
        * MPQFileHashB: The file's B hash as an integer. Mostly useless as well.
        
        * MPQFileHashPosition: The file's hash table position. Can be used as a unique key for that 
        particular file (in fact, a file's path is not a unique key, so this is the only unique key 
        for MPQ files).
        
        File information dictionaries may also contain the following key:
        
        * MPQFileName: The file's path inside the MPQ archive. Note that the path separator is \.
    @param lpszFileName  The MPQ file path of the file you wish information about. Must not be nil.
    @param wLocale  The file's locale code. See the MPQLocale enum in MPQSharedConstants.h for a list of valid values.
    @result  An NSDictionary object containing the file's information. Nil on failure or if the file is not found.
*/
- (NSDictionary *)fileInfoForFile:(NSString *)lpszFileName locale:(MPQLocale)wLocale;

/*! 
    @method fileInfoForFiles:locale:
	@abstract Returns the information dictionary for the specified files.
    @discussion File information dictionaries will always contain the following keys:
        
        * MPQFileLength: The size of the file in bytes as an integer.
        
        * MPQFileCompressedLength: The compressed size of the file in bytes as an integer.
        
        * MPQFileFlags: The file's flags as an integer. See the MPQFileFlag enum in 
        MPQSharedConstants.h for a list of valid bit values.
        
        * MPQFileLocale: The file's locale code as an integer. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
        
        * MPQFileHashA: The file's A hash as an integer. Mostly useless.
        
        * MPQFileHashB: The file's B hash as an integer. Mostly useless as well.
        
        * MPQFileHashPosition: The file's hash table position. Can be used as a unique key for that 
        particular file (in fact, a file's path is not a unique key, so this is the only unique key 
        for MPQ files).
        
        File information dictionaries may also contain the following key:
        
        * MPQFileName: The file's path inside the MPQ archive. Note that the path separator is \.
        
        This method simply call fileInfoForFile:locale: for each entry in fileArray and places the result in order
        inside an NSArray.
    @param fileArray  An array of MPQ file paths you wish information about. Must not be nil.
    @param wLocale  The locale to be used for every file in fileArray.
    @result  An NSArray of NSDictionary objects containing the files' information. Nil on failure.
*/
- (NSArray *)fileInfoForFiles:(NSArray *)fileArray locale:(MPQLocale)wLocale;

/*! 
    @method addFileWithPath:name:attributes:
	@abstract Adds the specified file to the MPQ archive.
    @discussion This method adds the file at path to the MPQ archive as lpszDestFile. Note that 
        the archive itself is not modified until writeToSave:atomically: is invoked. Note that this
        method simply calls addFileWithData:name:attributes:.
        
        The attribute dictionary is used to override the default import options. It may be nil, or 
        have the following keys:
        
        * MPQAddFlags: The file's flags. Default flags are MPQReplaceExisting and MPQCompress. 
        Note that MPQFileValid is automatically added to the flags, so you do not need 
        to specify it here. In order to compress the file, OR the value MPQCompress or 
        MPQCompressDiablo (not recommended) to the file flags. See the MPQFileFlag enum in 
        the MPQSharedConstants.h header for other flag values.
        
        * MPQAddQuality: If either compression flag is set, this is the compression quality that 
        will be used. Default is Z_DEFAULT_COMPRESSION for zlib compression and MPQWavQualityNone 
        for DCL/WAV compression. This flag is ignored for Diablo compression. If this flag is not 
        MPQWavQualityNone and DCL compression is used, then the file that is going to be added MUST
        BE A WAV FILE.
        
        * MPQAddLocale: The file's locale. Default value is MPQNeutral. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
        
        * MPQAddCompressior: The compressor that is going to be used for that particular file. 
        If this flag is not specified, the default compressor is used.
    @param path  The absolute path of the file on disk. Must not be nil.
    @param lpszDestFile  The MPQ file path. Note that the path separator MUST be \. Must not be nil.
    @param attributesDict  A NSDictionary containing file addition parameters. Can be nil.
    @result  YES on success or NO on failure.
*/
- (BOOL)addFileWithPath:(NSString *)path name:(NSString *)lpszDestFile attributes:(NSDictionary *)attributesDict;

/*! 
    @method addFileWithData:name:attributes:
	@abstract Adds the specified data to the MPQ archive.
    @discussion This method adds the given data to the MPQ archive as lpszDestFile. Note that 
        the archive itself is not modified until writeToSave:atomically: is invoked. As such, the 
        data will be retained until the archive has been saved.
        
        The attribute dictionary is used to override the default import options. It may be nil, or 
        have the following keys:
        
        * MPQAddFlags: The file's flags. Default flags are MPQReplaceExisting and MPQCompress. 
        Note that MPQFileValid is automatically added to the flags, so you do not need 
        to specify it here. In order to compress the file, OR the value MPQCompress or 
        MPQCompressDiablo (not recommended) to the file flags. See the MPQFileFlag enum in 
        the MPQSharedConstants.h header for other flag values.
        
        * MPQAddQuality: If either compression flag is set, this is the compression quality that 
        will be used. Default is Z_DEFAULT_COMPRESSION for zlib compression and MPQWavQualityNone 
        for DCL/WAV compression. This flag is ignored for Diablo compression. If this flag is not 
        MPQWavQualityNone and DCL compression is used, then the file that is going to be added MUST
        BE A WAV FILE.
        
        * MPQAddLocale: The file's locale. Default value is MPQNeutral. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
        
        * MPQAddCompressior: The compressor that is going to be used for that particular file. 
        If this flag is not specified, the default compressor is used.
    @param lpbySourceBuffer  The data to be added to the archive. Must not be nil.
    @param lpszDestFile  The MPQ file path. Note that the path separator MUST be \. Must not be nil.
    @param attributesDict  A NSDictionary containing file addition parameters. Can be nil.
    @result  YES on success or NO on failure.
*/
- (BOOL)addFileWithData:(NSData *)lpbySourceBuffer name:(NSString *)lpszDestFile attributes:(NSDictionary *)attributesDict;

/*! 
    @method deleteFile:
	@abstract Deletes the specified file from the MPQ archive.
    @discussion Note that when deleting files from MPQ archives, the file is simply marked
        as deleted. The data itself isn't removed from the archive file. That is to say, deleting 
        a file will not decrease the size of the archive. In order to delete unused file data from 
        an MPQ archive, use the compact method.
        
        WARNING: COMPACTION IS CURRENTLY NOT IMPLEMENTED.
        
        Note that this method simply calls deleteFile:locale: with MPQNeutral as the locale.
    @param lpszFileName  The path of the MPQ file to delete. Note that the path separator MUST be \. 
        Must not be nil.
    @result  YES on success or NO on failure.
*/
- (BOOL)deleteFile:(NSString *)lpszFileName;

/*! 
    @method deleteFile:locale:
	@abstract Deletes the specified file with the specified locale from the MPQ archive.
    @discussion Note that when deleting files from MPQ archives, the file is simply marked
        as deleted. The data itself isn't removed from the archive file. That is to say, deleting 
        a file will not decrease the size of the archive. In order to delete unused file data from 
        an MPQ archive, use the compact method.
        
        WARNING: COMPACTION IS CURRENTLY NOT IMPLEMENTED.
    @param lpszFileName  The path of the MPQ file to delete. Note that the path separator MUST be \. 
        Must not be nil.
    @param wLocale  The file's locale code. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
    @result  YES on success or NO on failure.
*/
- (BOOL)deleteFile:(NSString *)lpszFileName locale:(MPQLocale)wLocale;

/*! 
    @method renameFile:as:
	@abstract Renames the specified file as the new specified file.
    @discussion Renaming isn't as simple as it sounds. Indeed, since a file's hash table position 
        and file key are based on its file name (MPQ file path), renaming involves reading the 
        file's data, deleting the old file and adding a new one with the file's data under the 
        new name.
        
        Note that this method simply calls renameFile:as:locale: with MPQNeutral as the locale.
    @param lpszOldFileName  The path of the MPQ file to rename. Note that the path separator MUST be \. 
        Must not be nil.
    @param lpszNewFileName  The new path of the MPQ file. Note that the path separator MUST be \. 
        Must not be nil.
    @result  YES on success or NO on failure.
*/
- (BOOL)renameFile:(NSString *)lpszOldFileName as:(NSString *)lpszNewFileName;

/*! 
    @method renameFile:as:locale:
	@abstract Renames the specified file with the specified locale as the new specified file.
    @discussion Renaming isn't as simple as it sounds. Indeed, since a file's hash table position 
        and file key are based on its file name (MPQ file path), renaming involves reading the 
        file's data, deleting the old file and adding a new one with the file's data under the 
        new name.
    @param lpszOldFileName  The path of the MPQ file to rename. Note that the path separator MUST be \. 
        Must not be nil.
    @param lpszNewFileName  The new path of the MPQ file. Note that the path separator MUST be \. 
        Must not be nil.
    @param wLocale  The old and new file's locale code. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
	@result  YES on success or NO on failure.
*/
- (BOOL)renameFile:(NSString *)lpszOldFileName as:(NSString *)lpszNewFileName locale:(MPQLocale)wLocale;

/*! 
    @method openFile:
	@abstract Creates an MPQFile object for the specified file.
    @discussion MPQFile objects are useful for streaming a file's data. For one-time data reading, 
        you may use the dataForFile methods.
        
        Note that this method simply calls openFile:locale: with MPQNeutral as the locale.
    @param lpszFileName  The path of the MPQ file to open. Note that the path separator MUST be \. 
        Must not be nil.
    @result  An MPQFile instance on success or nil on failure.
*/
- (MPQFile *)openFile:(NSString *)lpszFileName;

/*! 
    @method openFile:locale:
	@abstract Creates an MPQFile object for the specified file.
    @discussion MPQFile objects are useful for streaming a file's data. For one-time data reading, 
        you may use the dataForFile methods.
    @param lpszFileName  The path of the MPQ file to open. Note that the path separator MUST be \. 
        Must not be nil.
    @param wLocale  The old and new file's locale code. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
    @result  An MPQFile instance on success or nil on failure.
*/
- (MPQFile *)openFile:(NSString *)lpszFileName locale:(MPQLocale)wLocale;

/*! 
    @method dataForFile:
	@abstract Returns the entire content of the specified file.
    @discussion Note that this method simply calls dataForFile:range:locale: with {0,0} as the range 
        and MPQNeutral as the locale.
        
        If you intend on streaming a file, you should not use the dataForFile methods, but rather 
        use the openFile methods to get a MPQFile object for the desired file.
    @param lpszFileName  The path of the MPQ file to read from. Note that the path separator MUST be \. 
        Must not be nil.
    @result  An NSData instance containing the requested data on success or nil on failure.
*/
- (NSData *)dataForFile:(NSString *)lpszFileName;

/*! 
    @method dataForFile:range:
	@abstract Returns the specified range of bytes from the specified file.
    @discussion Note that this method simply calls dataForFile:range:locale: with MPQNeutral as the locale.
        
        If you intend on streaming a file, you should not use the dataForFile methods, but rather 
        use the openFile methods to get a MPQFile object for the desired file.
    @param lpszFileName  The path of the MPQ file to read from. Note that the path separator MUST be \. 
        Must not be nil.
    @param dataRange  An NSRange struct specifying the range of bytes to read. Note that specifying 0 
        as the length will return all the data from the starting offset up until the end of file.
    @result  An NSData instance containing the requested data on success or nil on failure.
*/
- (NSData *)dataForFile:(NSString *)lpszFileName range:(NSRange)dataRange;

/*! 
    @method dataForFile:locale:
	@abstract Returns the entire content of the specified file with the specified locale.
    @discussion Note that this method simply calls dataForFile:range:locale: with {0,0} as the range.
        
        If you intend on streaming a file, you should not use the dataForFile methods, but rather 
        use the openFile methods to get a MPQFile object for the desired file.
    @param lpszFileName  The path of the MPQ file to read from. Note that the path separator MUST be \. 
        Must not be nil.
    @param wLocale  The file's locale code. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
    @result  An NSData instance containing the requested data on success or nil on failure.
*/
- (NSData *)dataForFile:(NSString *)lpszFileName locale:(MPQLocale)wLocale;

/*! 
    @method dataForFile:range:locale:
	@abstract Returns the specified range of bytes from the specified file with the specified locale.
    @discussion If you intend on streaming a file, you should not use the dataForFile methods, but rather 
        use the openFile methods to get a MPQFile object for the desired file.
    @param lpszFileName  The path of the MPQ file to read from. Note that the path separator MUST be \. 
        Must not be nil.
    @param dataRange  An NSRange struct specifying the range of bytes to read. Note that specifying 0 
        as the length will return all the data from the starting offset up until the end of file.
    @param wLocale  The file's locale code. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
	@result  An NSData instance containing the requested data on success or nil on failure.
*/
- (NSData *)dataForFile:(NSString *)lpszFileName range:(NSRange)dataRange locale:(MPQLocale)wLocale;

/*! 
    @method fileExists:
	@abstract Returns an array of locale codes for which the specified file exists.
	@discussion Files can have the same path in an MPQ archive so long as their locale is different.
    @param lpszFileName  The path of the MPQ file to search for. Note that the path separator MUST be \. 
        Must not be nil.
    @result  An NSArray instance containing the the list of locale codes for which the file exists, 
        or nil if the file does not exists.
*/
- (NSArray *)fileExists:(NSString *)lpszFileName;

/*! 
    @method fileExists:locale:
	@abstract Checks if a specified file exists in the MPQ archive for a specified locale.
	@discussion A file's path and locale are the two values necessary to uniquely identify a file in an MPQ archive.
    @param lpszFileName  The path of the MPQ file to search for. Note that the path separator MUST be \. 
        Must not be nil.
    @param wLocale  The file's locale code. See the MPQLocale enum in 
        MPQSharedConstants.h for a list of valid values.
    @result  YES if the file exists, or NO if it does not.
*/
- (BOOL)fileExists:(NSString *)lpszFileName locale:(MPQLocale)wLocale;

/*! 
    @method archiveInfo
	@abstract Returns the information dictionary for the MPQ archive.
    @discussion Archive information dictionaries will always contain the following keys:
        
        * MPQArchiveLength: The size of the archive on the disk in bytes as an integer.
        
        * MPQBlockSize: The size of file data blocks where the actual size in bytes
        of blocks is 512 << MPQBlockSize as an integer.
        
        * MPQNumberOfFiles: The number of used and deleted files in the archive as an integer.
        
        * MPQMaximumNumberOfFiles: The maximum number of files the archive may contain as an integer.
        
        * MPQNumberOfUsedFiles: The number of used files in the archive as an integer.
        
        * MPQArchiveOffset: MPQ archives may be embedded in other files with a 512 bytes padding.
        This is the offset of the MPQ archive within its container file in bytes as an integer.
        
        * MPQArchivePath: The path of the MPQ archive's container file. May be nil if the archive 
        was initialized without a file path and was not saved yet.
	@result  An NSDictionary object containing the archive's information on success or nil on failure.
*/
- (NSDictionary *)archiveInfo;

@end

/*!
    @category NSObject (MPQArchiveDelegate)
    @discussion The MPQArchiveDelegate delegate is used to report on the activity and status of 
        an MPQ archive and provide access management features to clients pertaining to all the 
        framework operations, such as adding, deleting, opening, etc.
*/
@interface NSObject (MPQArchiveDelegate)

/*!
    @method archiveShouldSave:
    @abstract This method is called immediately before saving an archive.
    @discussion Return YES in this delegate to perform the save operation, or NO to cancel 
        the save operation.
    @param archive The archive that is about to be saved.
	@result A response from the delegate.
*/
- (BOOL)archiveShouldSave:(MPQArchive *)archive;

/*!
    @method archiveWillSave:
    @abstract This method is called immediately before saving an archive but after
        archiveShouldSave: if it is implemented by the delegate.
	@discussion This message is always sent if the delegate implements it.
    @param archive The archive that is about to be saved.
*/
- (void)archiveWillSave:(MPQArchive *)archive;

/*!
    @method archiveDidSave:
    @abstract This method is called immediately after saving an archive.
    @discussion archiveDidSave: will not be sent to the delegate if the save operation failed.
    @param archive The archive that did save.
*/
- (void)archiveDidSave:(MPQArchive *)archive;

/*!
    @method archive:shouldAddFile:
    @abstract This method is called immediately before adding a file to an archive.
    @discussion This method is invoked when adding a file to the archive but prior to that file
        being actually compressed and written in the MPQ archive file. Return YES to permit the
        addition or NO to cancel it. Note that the new file will be accessible right after the
        addition method returns (if it was sucessfull). However, the data of the file will be 
        read from the original file on disk until the archive is saved, at which time the 
        delegate will be notified that the file is being compressed (if compression was 
        specified) and added to the MPQ archive file.
    @param archive The archive in which the file will be added.
    @param file The MPQ path of the file that is about to be added.
	@result A response from the delegate.
*/
- (BOOL)archive:(MPQArchive *)archive shouldAddFile:(NSString *)file;

/*!
    @method archive:willAddFile:
    @abstract This method is called immediately before the framework starts compressing a file
        into an archive.
    @discussion This method is called for each file pending for addition in the archive being 
        saved. This method is invoked after archiveWillSave: but prior to archiveDidSave:.
    @param archive The archive in which the file will be added.
    @param file The MPQ path of the file that is about to be added.
*/
- (void)archive:(MPQArchive *)archive willAddFile:(NSString *)file;

/*!
    @method archive:didAddFile:
    @abstract This method is called immediately after the framework compressed a file
        into an archive.
	@discussion This message is sent regardless if the file was compressed or not, encrypted or not.
    @param archive The archive in which the file was added.
    @param file The MPQ path of the file that was added.
*/
- (void)archive:(MPQArchive *)archive didAddFile:(NSString *)file;

/*!
    @method archive:shouldDeleteFile:
    @abstract This method is called immediately before deleting a file from an archive.
    @discussion Return YES to permit the addition or NO to cancel it.
    @param archive The archive in which file will be deleted.
    @param file The MPQ path of the file that is about to be deleted.
	@result A response from the delegate.
*/
- (BOOL)archive:(MPQArchive *)archive shouldDeleteFile:(NSString *)file;

/*!
    @method archive:willDeleteFile:
    @abstract This method is called immediately before the framework deletes a file
        from an archive.
    @discussion This method is invoked after archive:shouldDeleteFile:. Note that unlike file
        addition and renaming, archive:willDeleteFile: and archive:didDeleteFile: are not
        invoked when the archive is saved, but rather immediately before and after the delete file 
        method is invoked respectively. This is because no extra operation 
        is performed at save time for deletion, unlike addition and renaming. Note however that 
        deletion is undoable until the archive is saved.
    @param archive The archive from which file will be deleted.
    @param file The MPQ path of the file that is about to be deleted.
*/
- (void)archive:(MPQArchive *)archive willDeleteFile:(NSString *)file;

/*!
    @method archive:didDeleteFile:
    @abstract This method is called immediately after the framework deleted a file
        from an archive.
    @discussion Note that unlike file 
        addition and renaming, archive:willDeleteFile: and archive:didDeleteFile: are not 
        invoked when the archive is saved, but rather immediately before and after the delete file 
        method is invoked respectively. This is because no extra operation 
        is performed at save time for deletion, unlike addition and renaming. Note however that 
        deletion is undoable until the archive is saved.
    @param archive The archive from which file was deleted.
    @param file The MPQ path of the file that was deleted.
*/
- (void)archive:(MPQArchive *)archive didDeleteFile:(NSString *)file;

/*!
    @method archive:shouldRenameFile:as:
    @abstract This method is called immediately before renaming a file of an archive.
    @discussion This method is invoked when renaming a file of the archive but prior to that file
        being actually re-encrypted and re-written in the MPQ archive file. Return YES to permit the
        rename or NO to cancel it. Note that the file will be accessible right after the
        rename method returns (if it was sucessfull) as the new name. The 
        delegate will be notified that the file is being re-processed at save time.
    @param archive The archive in which the file to be renamed is stored.
    @param file The MPQ path of the file that is about to be renamed.
    @param newFile The new MPQ path of the file that is about to be renamed.
	@result A response from the delegate.
*/
- (BOOL)archive:(MPQArchive *)archive shouldRenameFile:(NSString *)file as:(NSString *)newFile;

/*!
    @method archive:willRenameFile:as:
    @abstract This method is called immediately before the framework starts renaming a file
        of an archive.
    @discussion This method is called for each file pending for renaming in the archive being 
        saved. This method is invoked after archiveWillSave: but prior to archiveDidSave:.
    @param archive The archive in which the file to be renamed is stored.
    @param file The MPQ path of the file that is about to be renamed.
    @param newFile The new MPQ path of the file that is about to be renamed.
*/
- (void)archive:(MPQArchive *)archive willRenameFile:(NSString *)file as:(NSString *)newFile;

/*!
    @method archive:didRenameFile:as:
    @abstract This method is called immediately after the framework renamed a file
        of an archive.
	@discussion It is safe to use the new name as soon as this message is sent.
    @param archive The archive in which the renamed file is stored.
    @param file The MPQ path of the file that was renamed.
    @param newFile The new MPQ path of the file that was renamed.
*/
- (void)archive:(MPQArchive *)archive didRenameFile:(NSString *)file as:(NSString *)newFile;

/*!
    @method archive:shouldOpenFile:
    @abstract This method is called immediately before opening a file of the archive.
    @discussion Return YES to permit the file to be opened or NO to prevent it.
    @param archive The archive containing the file that is about to be opened.
    @param file The MPQ path of the file that is about to be opened.
	@result A response from the delegate.
*/
- (BOOL)archive:(MPQArchive *)archive shouldOpenFile:(NSString *)file;

/*!
    @method archive:willOpenFile:
    @abstract This method is called immediately before opening a file of the archive.
    @discussion This method is invoked after archive:shouldOpenFile:.
    @param archive The archive containing the file that is about to be opened.
    @param file The MPQ path of the file that is about to be opened.
*/
- (void)archive:(MPQArchive *)archive willOpenFile:(NSString *)file;

/*!
    @method archive:didOpenFile:
    @abstract This method is called immediately after the framework opened a file
        from an archive.
    @discussion Will not be called if file opening is cancelled or failed.
    @param archive The archive from which file was deleted.
    @param file The MPQ path of the file that was deleted.
*/
- (void)archive:(MPQArchive *)archive didOpenFile:(NSString *)file;

@end

// MPQ structures
// umm...the whole structure definition can be here...
#pragma options align=packed
struct MPQHEADER {
        DWORD dwMPQID; 				// Must be "MPQ ", 0x4D50511A
        DWORD dwHeaderSize;         // Must be sizeof(MPQHEADER)
        DWORD dwMPQSize; 			// Total size of MPQ
        WORD wUnused0C;             // Must be 0
        WORD wBlockSize;            // Compression block size = 512 << wBlockSize
        DWORD dwHashTableOffset;
        DWORD dwFileTableOffset;
        DWORD dwHashTableSize;
        DWORD dwFileTableSize;
};
typedef struct MPQHEADER MPQHEADER;

struct FILETABLEENTRY {
        DWORD dwFileOffset;
        DWORD dwCompressedSize;
        DWORD dwFullSize;
        DWORD dwFlags;
};
typedef struct FILETABLEENTRY FILETABLEENTRY;

struct HASHTABLEENTRY {
        DWORD dwNameHashA;
        DWORD dwNameHashB;
        WORD wLocale;
        BYTE byPlatform;
        BYTE byUnused0B; 			// Must be 0
        DWORD dwFileTableIndex;
};
typedef struct HASHTABLEENTRY HASHTABLEENTRY;

// Internal structure for defered operations
struct internal_data_entry {
        HASHTABLEENTRY hash_entry;
        FILETABLEENTRY file_entry;
        unsigned file_key;
        NSString *name;
        unsigned rename_hash_index;
        NSData *data;
        NSDictionary *import_params;
        unsigned flags;
};
typedef struct internal_data_entry internal_data_entry;

struct ATTRIBUTES
{
    unsigned long magic; // Must be 0x64
    unsigned long attributes;
};
typedef struct ATTRIBUTES ATTRIBUTES;
#pragma options align=reset
